//////////////////////////////////////////////////////////////////////////////
//                                                                          
//  NAME: Marina Zlatkina
//  FILE: NonTypeArgs.C
//  PROJ: COKO
//  DATE: 8/2001                                                        
//
// SYNOPSIS
//  Class that represents the arguments to a property.
//  
//////////////////////////////////////////////////////////////////////////////
#include "NonTypeArgs.H"
#include "KOFVar.H"
#include "KOFUser.H"
#include <memory>
#include <strstream>

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name: NonTypeArgs
// Parameters:	(NonTypeList *, int)
//
// Constructor. Creates a NonTypeArgs object from a given NonTypeList object.
//         								    
//////////////////////////////////////////////////////////////////////////////
NonTypeArgs::NonTypeArgs( NonTypeList *l, int n ): list (l),num(n), success(false){};


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name: NonTypeArgs
// Parameters:	(int ...)
//
// Constructor. Creates a NonTypeArgs object from the NonType objects it
// receives.
//         								    
//////////////////////////////////////////////////////////////////////////////
NonTypeArgs:: NonTypeArgs( int a, ...){

  num=0;
  
  // start macros for reading variable number of arguments
  va_list args;
  va_start(args, a);

  // initialize list
  list = new NonTypeList();

  while (true){
    // read the next argument
    NonType *p = va_arg(args, NonType * );
   
    // if it is NULL, this is the last argument, leave loop.
    // otherwise, insert it into list.
    if (p==NULL)
	break;
    else
      list->insert(p);

    // increment the counter
    num++;
  }

  va_end(args);  
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name: build
// Parameters:	(CState *)
//
// Substitutes the variables in all the NonType objects in list with their
// values as stored in the current CState.
//         								    
//////////////////////////////////////////////////////////////////////////////

NonTypeArgs *NonTypeArgs::build(CState *s){
 
  // initialize the iterator  
  list->initIterator();
  
  NonTypeList *newList = new NonTypeList();
  NonType *current = NULL;

  // create a new CState containing a copy of the current NonType
  // (it doesn't matter what it is, it will be replaced anyway),
  // and the current environment.
  CState *state = new CState( s->Store()->Copy_kTree(), s->IEnv(), 0);
 
  // while we haven't reached the end of the list
  while ( current = list->getValue() ){
    
    // create a new CPattern from a copy of the current node in list
    std::auto_ptr<CPattern> pattern(new CPattern( current->Copy_kTree() ));

    // if it's a don't care variable, create a new FUser variable with
    // the regular expression matching a word.
    if ( current->Name()[ 0 ] == '_' )
      newList->insert( new FUser( "[a-zA-Z][a-z_A-Z0-9]*" ) );
    else{
      
      // call build method of pattern
      state = pattern->build(state);
      
      // if the previous call was successful, insert the new object.
      // otherwise,set the variable stack to NULL, perform clean up, and
      // return NULL.
      if ( state->ISuccess())
	newList->insert( state->Store()->Copy_kTree() );
      else{
	if( current->isChildVar(current))
	  newList->insert(current->Copy_kTree());
	else{
	  state->SetIEnv(NULL);
	  delete state;
	  delete current;
	  delete newList;
	  return new NonTypeArgs(new NonTypeList(), 0);
	}
      }
      
    }
  }

  // the variable stack is set to NULL, so it's not delete when state is.
  // (state and s have pointers that point to the same variable stack)
  state->SetIEnv(NULL);
  delete state;
  delete current;

  // return a NonTypeArgs object consisting of the new list just built.
  return new NonTypeArgs( newList, num );
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name: match
// Parameters:	(CState *, NonTypeArgs *)
//
// Match the arguments provided at runtime with the pattern arguments.
// This function uses the match method of class CPattern, so the matched
// variables and their values are stored in the CState.
//         								    
//////////////////////////////////////////////////////////////////////////////

CState *NonTypeArgs::match(CState *s, NonTypeArgs *a){
 
  // if the number of arguments in a is not the same as the number of
  // arguments in this object, set the success value to false.

  if ( num != a->numArgs() ){
    s->SetSuccess(F);
    return s;
  } 
  
  // initialize the iterator for the pattern
  list->initIterator();
  // retrieve the NonTypeList belonging to a
  NonTypeList *l = a->theList();
  // initialize it
  l->initIterator();

  CPattern *pattern=NULL;
  NonType *current = NULL;

  // create a new CState containing a copy of the current NonType
  // (it doesn't matter what it is, it will be replaced anyway),
  // and the current environment.
  CState *state = new CState(s->Store()->Copy_kTree(), s->IEnv(), 0);

  // while the iterator does not return NULL
  // (i.e., for every argument)
  while( current = list->getValue() ){   
   
    // set b to be a copy of the current iterator value for the NonTypeList
    // in a (the argument at runtime)
    NonType *b = l->getValue()->Copy_kTree();

    // create a new CPattern from a copy of the current iterator value
    // for list
    pattern = new CPattern( current->Copy_kTree() );
   
    // set the NonType in state to b
    state->SetStore( b );
   
    // call match function of the pattern. This call matches the pattern
    // argument to the actual argument.
    state = pattern->match( state );

    // if the match was not successful and we are not dealing with a variable,
    // , set the success value appropriately,
    // set the variable stack of state to point to NULL so it does not get 
    // deleted when state is (we want it to stay since state and s have
    // pointers to the same variable stack ). Return the original CState.
    
    if ( !state->ISuccess()){
      if ( !b->isChildVar(b) || !current->isChildVar(current) ){
	s->SetSuccess(F);
	state->SetIEnv( NULL );
	delete state;
	return s;
      }
      // if the match wasn't successful because we tried to match a variable
      // to a variable( the latter was a variable because we need to create a variable
      // binding for it),
      // insert onto the variable stack a pair containing the name
      // of the variable we have been trying to match and the variable we will create a 
      // binding for
      else{
	state->IInsert(current->Name(), new CIdent(NULL, b->Copy_kTree(), 0));
      }
    }
  }
  
  state->SetIEnv(NULL);

  // set sucess to true. 
  s->SetSuccess(T); 
  delete state;

  return s;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name: representation
// Parameters:	(void)
//
// Returns a char * representing the arguments.
//         								    
//////////////////////////////////////////////////////////////////////////////
char * NonTypeArgs::representation(){
  
  // initialize the iterator
  list->initIterator();
  
  NonType *current = list->getValue();  
  string s = " *\\( *";
  
  // if it's a variable - we need to create a binding for it, so
  // add a subexpression that will match an identifier
  if ( current->isChildVar(current))
    s.append( "([a-zA-Z][a-z_A-Z0-9]*)" );
  else
    s.append(current->Name());
  
  s.append(" *");
  
  // for every item in the list, append its representation
  while (current=list->getValue()){
    s.append(", *");
    
    // if it's a variable - we need to create a binding for it, so
    // add a subexpression that will match an identifier
    if ( current->isChildVar(current))
      s.append( "([a-zA-Z][a-z_A-Z0-9]*)" );
    else
      s.append(current->Name());

    s.append(" *");
  }  
 
  s.append("\\) *");  
  
  return ((char*)s.c_str());
}
//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name: print
// Parameters:	(void)
//
// prints out the represenation of every argument. This function is for
// debugging purposes.
//         								    
//////////////////////////////////////////////////////////////////////////////
void NonTypeArgs::print(){
  list->initIterator();
  NonType *current;

  while (current=list->getValue())
    std::cerr << current->representation() << "\n";
}



